/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "steghide.h"

static BUFFER *appendnewbuf (BUFFER *buflhead) ;

/* defines use or non-use of the noncnt buffer per file format
the array indices are the values of the FF_XXX constants */
int noncntbufuse[] = {	0,	/* FF_UNKNOWN ; not used */
						0,	/* FF_BMP */
						1,	/* FF_WAV */
						0,	/* FF_AU */	} ;

/* creates a CNTFILE structure that is used to save the stego data to disk */
CNTFILE *createstgfile (CNTFILE *cntfile, const char *stgfilename)
{
	CNTFILE *stgfile ;

	if ((stgfile = malloc (sizeof *stgfile)) == NULL)
		perr (ERR_MEMALLOC) ;
	
	if (strcmp (stgfilename, "-") == 0)
		stgfile->stream = stdout ;
	else {
		if ((stgfile->stream = fopen (stgfilename, "wb")) == NULL) {
			if (args_action == ACTN_EMBED)
				perr (ERR_STGOPEN) ;
			return NULL ;
		}
	}
	if ((stgfile->filename = malloc (strlen (stgfilename) + 1)) == NULL)
		perr (ERR_MEMALLOC) ;
	strcpy (stgfile->filename, stgfilename) ;
	stgfile->fileformat = cntfile->fileformat ;
	stgfile->headers = cntfile->headers ;
	stgfile->unsuphdr = cntfile->unsuphdr ;
	stgfile->unsuphdrlen = cntfile->unsuphdrlen ;
	stgfile->cntbuflhead = cntfile->cntbuflhead ;
	if (noncntbufuse[stgfile->fileformat])
		stgfile->noncntbuflhead = cntfile->noncntbuflhead ;
	else
		stgfile->noncntbuflhead = NULL ;

	return stgfile ;
}

/* creates a PLNFILE structure that is used to save the plaindata to disk */
PLNFILE *createplnfile (const char *plnfilename)
{
	PLNFILE *plnfile ;

	if ((plnfile = malloc (sizeof *plnfile)) == NULL)
		perr (ERR_MEMALLOC) ;

	if (strcmp (plnfilename, "-") == 0) {
		plnfile->stream = stdout ;
		plnfile->filename = NULL ;
	}
	else {
		if ((plnfile->stream = fopen (plnfilename, "wb")) == NULL)
			perr (ERR_PLNOPEN) ;

		if ((plnfile->filename = malloc (strlen (plnfilename) + 1)) == NULL)
			perr (ERR_MEMALLOC) ;
		strcpy (plnfile->filename, plnfilename) ;
	}

	plnfile->plnbuflhead = NULL ;	/* will be filled later by extractdata() */

	return plnfile ;
}

/* cleans up a CNTFILE structure (closes file, frees memory) */
void cleanupcntfile (CNTFILE *cntfile, int freesubstructs)
{
	if (freesubstructs == FSS_YES) {
		free (cntfile->headers) ;
		if (cntfile->unsuphdrlen != 0)
			free (cntfile->unsuphdr) ;
		buffree (cntfile->cntbuflhead) ;
		if (noncntbufuse[cntfile->fileformat])
			buffree (cntfile->noncntbuflhead) ;
	}
	if ((cntfile->filename != "-") && (cntfile->filename != "-"))
		fclose (cntfile->stream) ;
	free (cntfile->filename) ;
	free (cntfile) ;
}

/* clean up a PLNFILE structure (closes file, frees memory) */
void cleanupplnfile (PLNFILE *plnfile)
{
	if ((plnfile->filename != "-") && (plnfile->stream != NULL))
		fclose (plnfile->stream) ;
	if (plnfile->filename != NULL) ;
		free (plnfile->filename) ;
	if (plnfile->plnbuflhead != NULL)
		buffree (plnfile->plnbuflhead) ;
	free (plnfile) ;
}

/* initialises a linked list of BUFFER structures */
BUFFER *initbuflist()
{
	BUFFER *head, *end ;

	if ((head = malloc (sizeof *head)) == NULL)
		perr (ERR_MEMALLOC) ;

	if ((end = malloc (sizeof *end)) == NULL)
		perr (ERR_MEMALLOC) ;

	head->ptr = NULL ;
	head->bytepos = 0 ;
	head->bitpos = 0 ;
	head->nbytes = 0 ;
	head->nbytesalloc = 0 ;
	head->curbufnum = -1 ;
	head->nbufs = 0 ;
	head->prev = head ;
	head->next = end ;

	end->ptr = NULL ;
	end->bytepos = 0 ;
	end->bitpos = 0 ;
	end->nbytes = 0 ;
	end->nbytesalloc = 0 ;
	end->curbufnum = 0 ;
	end->nbufs = 0 ;
	end->prev = head ;
	end->next = end ;

	if (appendnewbuf (head) == NULL)
		return NULL ;

	return head ;
}

/* append a new BUFFER structure to a list */
static BUFFER *appendnewbuf (BUFFER *buflhead)
{
	BUFFER *newbuf, *lastbuf, *endbuf ;

	endbuf = buflhead ;
	while (endbuf != endbuf->next) {
		lastbuf = endbuf ;
		endbuf = lastbuf->next ;
	}

	if ((newbuf = malloc (sizeof *newbuf)) == NULL)
		perr (ERR_MEMALLOC) ;

	if ((newbuf->ptr = malloc (BUFSIZE)) == NULL)
		perr (ERR_MEMALLOC) ;
	newbuf->bytepos = 0 ;
	newbuf->bitpos = 0 ;
	newbuf->nbytes = 1 ;
	newbuf->nbytesalloc = BUFSIZE ;
	newbuf->curbufnum = 0 ;
	newbuf->nbufs = 0 ;
	newbuf->prev = lastbuf ;
	newbuf->next = endbuf ;

	lastbuf->next = newbuf ;
	endbuf->prev = newbuf ;

	buflhead->nbufs++ ;
	buflhead->curbufnum = buflhead->nbufs - 1 ;

	return newbuf ;
}

/* return a pointer to the current BUFFER structure in the given list */
BUFFER *getcurbuf (BUFFER *buflhead)
{
	BUFFER *curbuf ;
	int i ;

	curbuf = buflhead->next ;
	for (i = 0; i < buflhead->curbufnum; i++)
		curbuf = curbuf->next ;

	return curbuf ;
}

/* returns 1 if the "buffer pointer" points to the last byte of the data in buflhead */
int bufendofbuf (BUFFER *buflhead)
{
	int retval ;

	if ((buflhead->curbufnum == (buflhead->nbufs - 1)) && ((getcurbuf(buflhead)->bytepos + 1) == getcurbuf(buflhead)->nbytes))
		retval = 1 ;
	else
		retval = 0 ;

	return retval ;
}

/* returns the next 4 bytes in big endian encoding as ulong */
unsigned long bufget32_be (BUFFER *buflhead)
{
	unsigned long retval = 0x00000000 ;
	int byte, i ;

	for (i = 0 ; i < 4 ; i++)
		if ((byte = bufgetbyte (buflhead)) == ENDOFBUF)
			retval = (retval << 8) | 0x00 ;
		else
			retval = (retval << 8) | byte ;

	return retval ;
}

/* sets the next 4 bytes in big endian encoding to a ulong */
void bufset32_be (unsigned long val, BUFFER *buflhead)
{
	int i ;

	for (i = 0; i < 4 ; i++) {
		bufsetbyte ((val & 0xFF000000) >> 24, buflhead) ;
		val <<= 8 ;
	}

	return ;
}

/* returns the current byte from the given buffer list */
int bufgetbyte (BUFFER *buflhead)
{
	int retval ;

	if (bufendofbuf (buflhead))
		retval = ENDOFBUF ;
	else {
		BUFFER *curbuf ;
		unsigned char *bufptr ;

		curbuf = getcurbuf (buflhead) ;
		bufptr = curbuf->ptr ;
		retval = (int) bufptr[curbuf->bytepos] ;
		bufseek (buflhead, 1, BUFSEEK_CUR) ;
	}

	return retval ;
}

/* sets the current byte of the given buffer list to the given value */
void bufsetbyte (int byteval, BUFFER *buflhead)
{
	BUFFER *curbuf ;
	unsigned char *bufptr ;

	curbuf = getcurbuf (buflhead) ;

	bufptr = curbuf->ptr ;
	bufptr[curbuf->bytepos] = (unsigned char) byteval ;
	curbuf->bytepos++ ;
	curbuf->bitpos = 0 ;

	/* FIXME: DEBUG */
	if (curbuf->bytepos > curbuf->nbytes)
		printf ("this should not happen: bytepos > nbytes\n") ;

	if (curbuf->bytepos == curbuf->nbytes) {
		if (curbuf->nbytesalloc > curbuf->nbytes)
			curbuf->nbytes++ ;
		else {
			if (buflhead->curbufnum == (buflhead->nbufs - 1))
				curbuf = appendnewbuf (buflhead) ;
			else {
				curbuf->next->bytepos = 0 ;
				buflhead->curbufnum++ ;
			}
		}
	}

	return ;
}

/* get the current bit from the given buffer list */
int bufgetbit (BUFFER *buflhead)
{
	int retval ;

	if (bufendofbuf (buflhead))
		retval = ENDOFBUF ;
	else {
		BUFFER *curbuf ;
		unsigned char *bufptr ;

		curbuf = getcurbuf (buflhead) ;
		bufptr = curbuf->ptr ;
		retval = ((bufptr[curbuf->bytepos] & (1 << curbuf->bitpos)) >> curbuf->bitpos) ;

		if (curbuf->bitpos < 7)
			bufsetbitpos (bufgetbitpos (buflhead) + 1, buflhead) ;
		else {
			bufseek (buflhead, 1, BUFSEEK_CUR) ;
			bufsetbitpos (0, buflhead) ;
		}
	}

	return retval ;
}

/* sets the current bit of the given buffer list to the given value */
void bufsetbit (int bitval, BUFFER *buflhead)
{
	BUFFER *curbuf ;
	unsigned char *bufptr ;

	curbuf = getcurbuf (buflhead) ;

	bufptr = curbuf->ptr ;
	bufptr[curbuf->bytepos] = (bufptr[curbuf->bytepos] & (255 - (1 << curbuf->bitpos))) | ((bitval & 1) << curbuf->bitpos) ;

	if (curbuf->bitpos < 7)
		bufsetbitpos (bufgetbitpos (buflhead) + 1, buflhead) ;
	else {
		bufseek (buflhead, 1, BUFSEEK_CUR) ;
		bufsetbitpos (0, buflhead) ;
	}

	if (curbuf->bytepos == curbuf->nbytes) {
		if (curbuf->nbytesalloc > curbuf->nbytes)
			curbuf->nbytes++ ;
		else {
			curbuf = appendnewbuf (buflhead) ;
		}
	}

	return ;
}

/* sets the "bit pointer" to the given position */
void bufsetbitpos (int bitpos, BUFFER *buflhead) 
{
	if ((bitpos >= 0) && (bitpos <= 7))
		getcurbuf(buflhead)->bitpos = bitpos ;

	return ;
}

/* returns value of the "bit pointer" */
int bufgetbitpos (BUFFER *buflhead)
{
	return getcurbuf(buflhead)->bitpos ;
}

/* sets "pointer" of buffer list to new value */
int bufseek (BUFFER *buflhead, long offset, int whence)
{
	BUFFER *curbuf ;
	int retval = 0 ;

	curbuf = getcurbuf (buflhead) ;

	switch (whence) {
		case BUFSEEK_SET:
		offset = offset - buftell (buflhead) ;
		break ;

		case BUFSEEK_END:
		offset = (bufgetdatalen (buflhead) - 1 + offset) - buftell (buflhead) ;
		break ;
	}

	while (offset > 0) {
		curbuf->bitpos = 0 ;
		while ((curbuf->bytepos < curbuf->nbytes) && (offset > 0)) {
			curbuf->bytepos++ ;
			offset-- ;
		}

		if ((buflhead->curbufnum == (buflhead->nbufs - 1)) && ((curbuf->bytepos + 1) >= curbuf->nbytes)) {
			retval = ENDOFBUF ;
			break ;
		}

		if ((buflhead->curbufnum < buflhead->nbufs - 1) && (curbuf->bytepos == curbuf->nbytes)) {
			curbuf = curbuf->next ;
			curbuf->bytepos = 0 ;
			buflhead->curbufnum++ ;
		}
	}

	while (offset < 0) {
		curbuf->bitpos = 0 ;
		while ((curbuf->bytepos >= 0) && (offset < 0)) {
			curbuf->bytepos-- ;
			offset++ ;
		}

		if (curbuf->bytepos == -1) {
			if (buflhead->curbufnum == 0) {
				retval = ENDOFBUF ;
				curbuf->bytepos = 0 ;
				break ;
			}
			else {
				curbuf = curbuf->prev ;
				curbuf->bytepos-- ;
				buflhead->curbufnum-- ;
			}
		}
	}

	return retval ;
}

/* returns value of the "pointer" */
long buftell (BUFFER *buflhead)
{
	BUFFER *curbuf ;
	unsigned long pos = 0 ;
	int i ;

	curbuf = buflhead->next ;
	for (i = 0; i < buflhead->curbufnum; i++) {
		pos += curbuf->nbytes ;
		curbuf = curbuf->next ;
	}

	pos += curbuf->bytepos ;

	return pos ;
}

/* returns the number of bytes that are saved in the buffer list */
long bufgetdatalen (BUFFER *buflhead)
{
	BUFFER *curbuf ;
	long bytes = 0 ;

	curbuf = buflhead->next ;
	while (curbuf != curbuf->next) {
		bytes += curbuf->nbytes ;
		curbuf = curbuf->next ;
	}

	/* there always is one unused byte at the end */
	return bytes - 1 ;
}

/* frees the memory that is occupied by a buffer list */
void buffree (BUFFER *buflhead)
{
	BUFFER *curbuf, *nextbuf ;

	curbuf = buflhead->next ;
	free (buflhead) ;

	while (curbuf != curbuf->next) {
		nextbuf = curbuf->next ;
		free (curbuf->ptr) ;
		free (curbuf) ;
		curbuf = nextbuf ;
	}

	free (curbuf) ;
	return ;
}
