/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "steghide.h"
#include "blowfish.h"

#define BF_ROUNDS	16

unsigned long bf_P[18] ;
unsigned long bf_S[4][256] ;

static void bf_init (unsigned char *key, int keylen) ;
static void bf_encrypt (PLNFILE *plnfile) ;
static void bf_decrypt (PLNFILE *plnfile) ;
static void bf_encblock (unsigned long *xl, unsigned long *xr) ;
static void bf_decblock (unsigned long *xl, unsigned long *xr) ;
static unsigned long bf_f (unsigned long x) ;
static void swap (unsigned long *x, unsigned long *y) ;

void encrypt_data (PLNFILE *plnfile, int encryption, char *passphrase)
{
	unsigned char *key ;

	switch (encryption) {
		case ENC_BLOWFISH:
		key = md5_generatekey (passphrase) ;	
		bf_init (key, 16) ;
		bf_encrypt (plnfile) ;
		free (key) ;
		break ;
	}

	return ;
}

void decrypt_data (PLNFILE *plnfile, int encryption, char *passphrase)
{
	unsigned char *key ;

	switch (encryption) {
		case ENC_BLOWFISH:
		key = md5_generatekey (passphrase) ;
		bf_init (key, 16) ;
		bf_decrypt (plnfile) ;
		free (key) ;
		break ;
	}

	return ;
}

void encrypt_sthdr (unsigned char *bytes, int len, int encryption, char *passphrase)
{
	unsigned char *key ;
	int i ;

	switch (encryption) {
		case ENC_BLOWFISH:
		if (len < STHDR_BLOCKS_BLOWFISH * 8)
			while (len % 8 != 0)
				bytes[len++] = '\0' ;

		key = md5_generatekey (passphrase) ;
		bf_init (key, 16) ;
		for (i = 0 ; i < len ; i += 8) {
			unsigned long xl, xr ;

			cp32uc2ul_be (&xl, bytes + i) ;
			cp32uc2ul_be (&xr, bytes + i + 4) ;
			bf_encblock (&xl, &xr) ;
			cp32ul2uc_be (bytes + i, xl) ;
			cp32ul2uc_be (bytes + i + 4, xr) ;
		}
		free (key) ;
	}

	return ;
}

void decrypt_sthdr (unsigned char *bytes, int encryption, char *passphrase)
{
	unsigned char *key ;
	int i ;

	switch (encryption) {
		case ENC_BLOWFISH:
		key = md5_generatekey (passphrase) ;
		bf_init (key, 16) ;
		for (i = 0 ; i < STHDR_BLOCKS_BLOWFISH * 8 ; i += 8) {
			unsigned long xl, xr ;

			cp32uc2ul_be (&xl, bytes + i) ;
			cp32uc2ul_be (&xr, bytes + i + 4) ;
			bf_decblock (&xl, &xr) ;
			cp32ul2uc_be (bytes + i, xl) ;
			cp32ul2uc_be (bytes +i + 4, xr) ;
		}
		free (key) ;
	}

	return ;
}

/* initialises the P array and the S boxes */
static void bf_init (unsigned char *key, int keylen)
{
	int i, j, k ;
	unsigned long ulkey, xl = 0x00000000, xr = 0x00000000 ;

	for (i = 0 ; i < BF_ROUNDS + 2 ; i++)
		bf_P[i] = bf_P_init[i] ;

	for (i = 0 ; i < 4 ; i++)
		for (j = 0 ; j < 256 ; j++)
			bf_S[i][j] = bf_S_init[i][j] ;

	j = 0 ;
	for (i = 0 ; i < BF_ROUNDS + 2 ; i++) {
		ulkey = 0x00000000 ;
		for (k = 0 ; k < 4 ; k++) {
			ulkey = (ulkey << 8) | key[j] ;
			if (++j >= keylen)
				j = 0 ;	
		}
		bf_P[i] = bf_P[i] ^ ulkey ;
	}

	for (i = 0 ; i < BF_ROUNDS + 2 ; i += 2) {
		bf_encblock (&xl, &xr) ;
		bf_P[i] = xl ;
		bf_P[i + 1] = xr ;
	}

	for (i = 0 ; i < 4 ; i++)
		for (j = 0 ; j < 256 ; j += 2) {
			bf_encblock (&xl, &xr) ;
			bf_S[i][j] = xl ;
			bf_S[i][j + 1] = xr ;
		}
}

/* encrypts the data in the plnfile structure */
static void bf_encrypt (PLNFILE *plnfile)
{
	long pos ;

	while (!bufendofbuf (plnfile->plnbuflhead)) {
		unsigned long xl, xr ;

		pos = buftell (plnfile->plnbuflhead) ;
		xl = bufget32_be (plnfile->plnbuflhead) ;
		xr = bufget32_be (plnfile->plnbuflhead) ;
		bf_encblock (&xl, &xr) ;
		bufseek (plnfile->plnbuflhead, pos, SEEK_SET) ;
		bufset32_be (xl, plnfile->plnbuflhead) ;
		bufset32_be (xr, plnfile->plnbuflhead) ;
	}

	bufseek (plnfile->plnbuflhead, 0, SEEK_SET) ;

	return ;
}

static void bf_decrypt (PLNFILE *plnfile)
{
	long pos ;
	int reallen, paddedlen ;

	while (!bufendofbuf (plnfile->plnbuflhead)) {
		unsigned long xl, xr ;

		pos = buftell (plnfile->plnbuflhead) ;
		xl = bufget32_be (plnfile->plnbuflhead) ;
		xr = bufget32_be (plnfile->plnbuflhead) ;
		bf_decblock (&xl, &xr) ;
		bufseek (plnfile->plnbuflhead, pos, SEEK_SET) ;
		bufset32_be (xl, plnfile->plnbuflhead) ;
		bufset32_be (xr, plnfile->plnbuflhead) ;
	}

	reallen = sthdr.nbytesplain ;
	paddedlen = bufgetdatalen (plnfile->plnbuflhead) ;
	if (paddedlen != reallen) {
		BUFFER *curbuf ;

		curbuf = getcurbuf (plnfile->plnbuflhead) ;
		if (curbuf->nbytes >= paddedlen - reallen)
			curbuf->nbytes -= (paddedlen - reallen) ;
		else {
			curbuf->prev->nbytes -= ((paddedlen - reallen) - curbuf->nbytes) ;

			curbuf->prev->next = curbuf->next ;
			curbuf->next->prev = curbuf->prev ;
			free (curbuf->ptr) ;
			free (curbuf) ;
		}
	}

	bufseek (plnfile->plnbuflhead, 0, SEEK_SET) ;

	return ;
}

/* encrypts one block of data */
static void bf_encblock (unsigned long *xl, unsigned long *xr)
{
	unsigned long l, r ;
	int i ;

	l = *xl ;
	r = *xr ;

	for (i = 0 ; i < BF_ROUNDS ; i++) {
		l = l ^ bf_P[i] ;
		r = r ^ bf_f (l) ;

		swap (&l, &r) ;
	}

	swap (&l, &r) ;

	r = r ^ bf_P[BF_ROUNDS] ;
	l = l ^ bf_P[BF_ROUNDS + 1] ;

	*xl = l ;
	*xr = r ;
}

/* decrypts one block of data */
static void bf_decblock (unsigned long *xl, unsigned long *xr)
{
	unsigned long l, r ;
	int i ;

	l = *xl ;
	r = *xr ;

	for (i = BF_ROUNDS + 1 ; i > 1 ; i--) {
		l = l ^ bf_P[i] ;
		r = r ^ bf_f (l) ;

		swap (&l, &r) ;
	}

	swap (&l, &r) ;

	r = r ^ bf_P[1] ;
	l = l ^ bf_P[0] ;

	*xl = l ;
	*xr = r ;
}

/* the blowfish function F */
static unsigned long bf_f (unsigned long x)
{
	unsigned char a, b, c, d ;

	d = x & 0xFF ;
	c = (x >> 8) & 0xFF ;
	b = (x >> 16) & 0xFF ;
	a = (x >> 24) & 0xFF ;

	return (((bf_S[0][a] + bf_S[1][b]) ^ bf_S[2][c]) + bf_S[3][d]) ;
}

static void swap (unsigned long *x, unsigned long *y)
{
	unsigned long t ;

	t = *x ;
	*x = *y ;
	*y = t ;
}
