/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "steghide.h"

static void bmp_readheaders (CNTFILE *file) ;
static void bmp_writeheaders (CNTFILE *file) ;
static void wav_readheaders (CNTFILE *file, unsigned long rifflen) ;
static void wav_writeheaders (CNTFILE *file) ;
static void wav_getchhdr (FILE *file, CHUNKHEADER *chhdr) ;
static void wav_putchhdr (FILE *file, CHUNKHEADER *chhdr) ;
static void au_readheaders (CNTFILE *file) ;
static void au_writeheaders (CNTFILE *file) ;

/* auto-detects file format while reading headers into file->headers */
void readheaders (CNTFILE *file)
{
	char buf[4] ;
	int i ;

	for (i = 0 ; i < 2 ; i++)
		buf[i] = getc (file->stream) ;

	if (strncmp ("BM", buf, 2) == 0) {
		file->fileformat = FF_BMP ;
		bmp_readheaders (file) ;
	}
	else {
		for (i = 2 ; i < 4 ; i++)
			buf[i] = getc (file->stream) ;

		if (strncmp (".snd", buf, 4) == 0) {
			file->fileformat = FF_AU ;
			au_readheaders (file) ;
		}
		else if (strncmp ("RIFF", buf, 4) == 0) {
			unsigned long rifflen ;

			rifflen = read32_le (file->stream) ;
			for (i = 0 ; i < 4 ; i++)
				buf[i] = getc (file->stream) ;
			if (strncmp ("WAVE", buf, 4) == 0) {
				file->fileformat = FF_WAV ;
				wav_readheaders (file, rifflen) ;
			}
			else {
				if (args_action == ACTN_EMBED)
					perr (ERR_CNTUNKNOWNFFORMAT) ;
				else if (args_action == ACTN_EXTRACT)
					perr (ERR_STGUNKNOWNFFORMAT) ;
			}
		}
		else {
			if (args_action == ACTN_EMBED)
				perr (ERR_CNTUNKNOWNFFORMAT) ;
			else if (args_action == ACTN_EXTRACT)
				perr (ERR_STGUNKNOWNFFORMAT) ;
		}
	}

	return ;
}


/* reads the headers of a bmp file from disk */
static void bmp_readheaders (CNTFILE *file)
{
	file->headers->bmp.bmfh.bfType = BMP_ID_BM ;
	file->headers->bmp.bmfh.bfSize = read32_le (file->stream) ;
	file->headers->bmp.bmfh.bfReserved1 = read16_le (file->stream) ;
	file->headers->bmp.bmfh.bfReserved2 = read16_le (file->stream) ;
	file->headers->bmp.bmfh.bfOffBits = read32_le (file->stream) ;
	
	file->headers->bmp.bmxh.Size = read32_le (file->stream) ;
	switch (file->headers->bmp.bmxh.Size) {
		case BMP_SIZE_BMINFOHEADER:
		file->headers->bmp.bmxh.Width = read32_le (file->stream) ;
		file->headers->bmp.bmxh.Height = read32_le (file->stream) ;
		file->headers->bmp.bmxh.Planes = read16_le (file->stream) ;
		file->headers->bmp.bmxh.BitCount = read16_le (file->stream) ;
		if ((file->headers->bmp.bmxh.Compression = read32_le (file->stream)) != BMP_BI_RGB) {
			if (args_action == ACTN_EMBED)
				perr (ERR_CNTBMPWITHCOMPR) ;
			else if (args_action == ACTN_EXTRACT)
				perr (ERR_STGBMPWITHCOMPR) ;
		}
		file->headers->bmp.bmxh.SizeImage = read32_le (file->stream) ;
		file->headers->bmp.bmxh.XPelsPerMeter = read32_le (file->stream) ;
		file->headers->bmp.bmxh.YPelsPerMeter = read32_le (file->stream) ;
		file->headers->bmp.bmxh.ClrUsed = read32_le (file->stream) ;
		file->headers->bmp.bmxh.ClrImportant = read32_le (file->stream) ;
		break ;

		case BMP_SIZE_BMCOREHEADER:
		file->headers->bmp.bmxh.Width = read16_le (file->stream) ;
		file->headers->bmp.bmxh.Height = read16_le (file->stream) ;
		file->headers->bmp.bmxh.Planes = read16_le (file->stream) ;
		file->headers->bmp.bmxh.BitCount = read16_le (file->stream) ;
		break ;

		default:
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTUNSUPBMPFILE) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGUNSUPBMPFILE) ;
	}

	if (file->headers->bmp.bmxh.BitCount == 24) {
		file->unsuphdr = NULL ;
		file->unsuphdrlen = 0 ;
	}
	else {
		unsigned char *ptrunsuphdr ;
		int i ;

		file->unsuphdrlen = file->headers->bmp.bmfh.bfOffBits - (BMP_SIZE_BMFILEHEADER + file->headers->bmp.bmxh.Size) ;

		if ((file->unsuphdr = malloc (file->unsuphdrlen)) == NULL)
			perr (ERR_MEMALLOC) ;

		ptrunsuphdr = file->unsuphdr ;
		for (i = 0 ; i < file->unsuphdrlen ; i++)
			ptrunsuphdr[i] = (unsigned char) getc (file->stream) ;
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTREAD) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGREAD) ;
	}

	return ;
}

/* writes the headers of a bmp file to disk */
static void bmp_writeheaders (CNTFILE *file)
{
	write16_le (file->stream, file->headers->bmp.bmfh.bfType) ;
	write32_le (file->stream, file->headers->bmp.bmfh.bfSize) ;
	write16_le (file->stream, file->headers->bmp.bmfh.bfReserved1) ;
	write16_le (file->stream, file->headers->bmp.bmfh.bfReserved2) ;
	write32_le (file->stream, file->headers->bmp.bmfh.bfOffBits) ;

	switch (file->headers->bmp.bmxh.Size) {
		case BMP_SIZE_BMINFOHEADER:
		write32_le (file->stream, file->headers->bmp.bmxh.Size) ;
		write32_le (file->stream, file->headers->bmp.bmxh.Width) ;
		write32_le (file->stream, file->headers->bmp.bmxh.Height) ;
		write16_le (file->stream, file->headers->bmp.bmxh.Planes) ;
		write16_le (file->stream, file->headers->bmp.bmxh.BitCount) ;
		write32_le (file->stream, file->headers->bmp.bmxh.Compression) ;
		write32_le (file->stream, file->headers->bmp.bmxh.SizeImage) ;
		write32_le (file->stream, file->headers->bmp.bmxh.XPelsPerMeter) ;
		write32_le (file->stream, file->headers->bmp.bmxh.YPelsPerMeter) ;
		write32_le (file->stream, file->headers->bmp.bmxh.ClrUsed) ;
		write32_le (file->stream, file->headers->bmp.bmxh.ClrImportant) ;
		break ;

		case BMP_SIZE_BMCOREHEADER:
		write32_le (file->stream, file->headers->bmp.bmxh.Size) ;
		write16_le (file->stream, file->headers->bmp.bmxh.Width) ;
		write16_le (file->stream, file->headers->bmp.bmxh.Height) ;
		write16_le (file->stream, file->headers->bmp.bmxh.Planes) ;
		write16_le (file->stream, file->headers->bmp.bmxh.BitCount) ;
		break ;
	}

	if (file->unsuphdrlen != 0) {
		unsigned char *ptrunsuphdr = file->unsuphdr ;
		int i ;

		for (i = 0 ; i < file->unsuphdrlen ; i++)
			putc ((int) ptrunsuphdr[i], file->stream) ;
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_STGWRITE) ;
	}

	return ;
}

/* reads a bmp file from disk into a CNTFILE structure */
void bmp_readfile (CNTFILE *file)
{
	long posinline = 0, line = 0 ;

	while (line < file->headers->bmp.bmxh.Height) {
		while (posinline < file->headers->bmp.bmxh.BitCount * file->headers->bmp.bmxh.Width / 8) {
			bufsetbyte (getc (file->stream), file->cntbuflhead) ;
			posinline++ ;
		}

		while (posinline++ % 4 != 0)
			getc (file->stream) ;

		line++ ;
		posinline = 0 ;
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTREAD) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGREAD) ;
	}
	
	return ;
}

/* writes a bmp file from a CNTFILE structure to disk */
void bmp_writefile (CNTFILE *file)
{
	long posinline = 0, line = 0;

	bmp_writeheaders (file) ;

	while (line < file->headers->bmp.bmxh.Height) {
		while (posinline < file->headers->bmp.bmxh.BitCount * file->headers->bmp.bmxh.Width / 8) {
			putc (bufgetbyte (file->cntbuflhead), file->stream) ;
			posinline++ ;
		}

		while (posinline++ % 4 != 0)
			putc (0, file->stream) ;

		line++ ;
		posinline = 0 ;
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_STGWRITE) ;
	}

	return ;
}

/* reads the headers of a wav file from disk */
static void wav_readheaders (CNTFILE *file, unsigned long rifflen)
{
	CHUNKHEADER tmpchhdr ;
	int i ;

	strcpy (file->headers->wav.riffchhdr.id, "RIFF") ;
	file->headers->wav.riffchhdr.len = rifflen ;
	strcpy (file->headers->wav.id_wave, "WAVE") ;

	wav_getchhdr (file->stream, &file->headers->wav.fmtchhdr) ;
	if ((file->headers->wav.fmtch.FormatTag = read16_le (file->stream)) != WAV_FORMAT_PCM) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTUNSUPWAVFILE) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGUNSUPWAVFILE) ;
	}
	file->headers->wav.fmtch.Channels = read16_le (file->stream) ;
	file->headers->wav.fmtch.SamplesPerSec = read32_le (file->stream) ;
	file->headers->wav.fmtch.AvgBytesPerSec = read32_le (file->stream) ;
	file->headers->wav.fmtch.BlockAlign = read16_le (file->stream) ;
	/* if a number other than a multiple of 8 is used, we cannot hide data,
	   because the least significant bits are always set to zero */
	if ((file->headers->wav.fmtch.BitsPerSample = read16_le (file->stream)) % 8 != 0) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTBPSNOTMULTOF8) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGBPSNOTMULTOF8) ;
	}

	file->unsuphdr = NULL ;
	file->unsuphdrlen = 0 ;
	wav_getchhdr (file->stream, &tmpchhdr) ;
	while (strncmp (tmpchhdr.id, "data", 4) != 0) {
		unsigned char *ptrunsuphdr ;

		if ((file->unsuphdr = realloc (file->unsuphdr, (file->unsuphdrlen + 8 + tmpchhdr.len))) == NULL)
			perr (ERR_MEMALLOC) ;

		ptrunsuphdr = file->unsuphdr ;
		for (i = 0; i < 4; i++)
			ptrunsuphdr[file->unsuphdrlen++] = (unsigned char) tmpchhdr.id[i] ;
		cp32ul2uc_le (&ptrunsuphdr[file->unsuphdrlen], tmpchhdr.len) ;
		file->unsuphdrlen += 4 ;

		for (i = 0; i < tmpchhdr.len; i++)
			ptrunsuphdr[file->unsuphdrlen++] = (unsigned char) getc (file->stream) ;

		wav_getchhdr (file->stream, &tmpchhdr) ;
	}

	strncpy (file->headers->wav.datachhdr.id, tmpchhdr.id, 4) ;
	file->headers->wav.datachhdr.len = tmpchhdr.len ;

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTREAD) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGREAD) ;
	}

	return ;
}

/* writes the headers of a wav file to disk */
static void wav_writeheaders (CNTFILE *file)
{
	int i ;

	wav_putchhdr (file->stream, &file->headers->wav.riffchhdr) ;
	for (i = 0; i < 4; i++)
		putc ((int) file->headers->wav.id_wave[i], file->stream) ;

	wav_putchhdr (file->stream, &file->headers->wav.fmtchhdr) ;
	write16_le (file->stream, file->headers->wav.fmtch.FormatTag) ;
	write16_le (file->stream, file->headers->wav.fmtch.Channels) ;
	write32_le (file->stream, file->headers->wav.fmtch.SamplesPerSec) ;
	write32_le (file->stream, file->headers->wav.fmtch.AvgBytesPerSec) ;
	write16_le (file->stream, file->headers->wav.fmtch.BlockAlign) ;
	write16_le (file->stream, file->headers->wav.fmtch.BitsPerSample) ;

	if (file->unsuphdrlen > 0) {
		unsigned char *ptrunsuphdr = file->unsuphdr ;
		for (i = 0; i < file->unsuphdrlen; i++)
			putc ((int) ptrunsuphdr[i], file->stream) ;
	}

	wav_putchhdr (file->stream, &file->headers->wav.datachhdr) ;

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_STGWRITE) ;
	}

	return ;
}

/* reads a wav file from disk into a CNTFILE structure */
void wav_readfile (CNTFILE *file)
{
	int numnoncntbytes, pos = 0, i ;

	numnoncntbytes = (file->headers->wav.fmtch.BitsPerSample / 8) - 1 ;

	while (pos < file->headers->wav.datachhdr.len) {
		bufsetbyte (getc (file->stream), file->cntbuflhead) ;
		pos++ ;

		for (i = 0; i < numnoncntbytes; i++) {
			bufsetbyte (getc (file->stream), file->noncntbuflhead) ;
			pos++ ;
		}
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTREAD) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGREAD) ;
	}

	return ;
}

/* writes a wav file from a CNTFILE structure to disk */
void wav_writefile (CNTFILE *file)
{
	int numnoncntbytes, pos = 0, i ;

	wav_writeheaders (file) ;

	numnoncntbytes = (file->headers->wav.fmtch.BitsPerSample / 8) - 1 ;

	while (pos < file->headers->wav.datachhdr.len) {
		putc (bufgetbyte (file->cntbuflhead), file->stream) ;
		pos++ ;

		for (i = 0; i < numnoncntbytes; i++) {
			putc (bufgetbyte (file->noncntbuflhead), file->stream) ;
			pos++ ;
		}
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_STGWRITE) ;
	}

	return ;
}

/* reads a wav chunk header from file and puts it indo chhdr */
static void wav_getchhdr (FILE *file, CHUNKHEADER *chhdr)
{
	chhdr->id[0] = getc (file) ;
	chhdr->id[1] = getc (file) ;
	chhdr->id[2] = getc (file) ;
	chhdr->id[3] = getc (file) ;
	chhdr->len = read32_le (file) ;
}

/* writes a wav chunk header from chhdr to file */
static void wav_putchhdr (FILE *file, CHUNKHEADER *chhdr)
{
	putc (chhdr->id[0], file) ;
	putc (chhdr->id[1], file) ;
	putc (chhdr->id[2], file) ;
	putc (chhdr->id[3], file) ;
	write32_le (file, chhdr->len) ;
}

static void au_readheaders (CNTFILE *file)
{
	strcpy (file->headers->au.id, ".snd") ;
	file->headers->au.offset = read32_be (file->stream) ;
	file->headers->au.size = read32_be (file->stream) ;
	file->headers->au.encoding = read32_be (file->stream) ;
	file->headers->au.samplerate = read32_be (file->stream) ;
	file->headers->au.channels = read32_be (file->stream) ;

	if ((file->unsuphdrlen = (file->headers->au.offset - AU_SIZE_HEADER)) != 0) {
		unsigned char *ptrunsuphdr ;
		int i ;

		if ((file->unsuphdr = malloc (file->unsuphdrlen)) == NULL)
			perr (ERR_MEMALLOC) ;

		ptrunsuphdr = file->unsuphdr ;
		for (i = 0 ; i < file->unsuphdrlen ; i++)
			ptrunsuphdr[i] = (unsigned char) getc (file->stream) ;
	}
		
	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTREAD) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGREAD) ;
	}

	return ;
}

static void au_writeheaders (CNTFILE *file)
{
	putc (file->headers->au.id[0], file->stream) ;
	putc (file->headers->au.id[1], file->stream) ;
	putc (file->headers->au.id[2], file->stream) ;
	putc (file->headers->au.id[3], file->stream) ;
	write32_be (file->stream, file->headers->au.offset) ;
	write32_be (file->stream, file->headers->au.size) ;
	write32_be (file->stream, file->headers->au.encoding) ;
	write32_be (file->stream, file->headers->au.samplerate) ;
	write32_be (file->stream, file->headers->au.channels) ;

	if (file->unsuphdrlen != 0) {
		unsigned char *ptrunsuphdr = file->unsuphdr ;
		int i ;

		for (i = 0 ; i < file->unsuphdrlen ; i++)
			putc ((int) ptrunsuphdr[i], file->stream) ;
	}

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_STGWRITE) ;
	}

	return ;
}

void au_readfile (CNTFILE *file)
{
	int c ;

	while ((c = getc (file->stream)) != EOF)
		bufsetbyte (c, file->cntbuflhead) ;

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_CNTREAD) ;
		else if (args_action == ACTN_EXTRACT)
			perr (ERR_STGREAD) ;
	}

	return ;
}

void au_writefile (CNTFILE *file)
{
	int c ;

	au_writeheaders (file) ;

	while ((c = bufgetbyte (file->cntbuflhead)) != ENDOFBUF)
		putc (c, file->stream) ;

	if (ferror (file->stream)) {
		if (args_action == ACTN_EMBED)
			perr (ERR_STGWRITE) ;
	}

	return ;
}
