/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "steghide.h"

/* reads a container data file into a CNTFILE structure */
CNTFILE *readcntfile (const char *filename)
{
	CNTFILE *cntfile ;

	/* fill cntfile structure */
	if ((cntfile = malloc (sizeof *cntfile)) == NULL)
		perr (ERR_MEMALLOC) ;

	if (strcmp (filename, "-") == 0)
		cntfile->stream = stdin ;
	else {
		if ((cntfile->stream = fopen (filename, "rb")) == NULL) {
			if (args_action == ACTN_EMBED)
				perr (ERR_CNTOPEN) ;
			else if (args_action == ACTN_EXTRACT)
				perr (ERR_STGOPEN) ;
			return NULL ;
		}
	}

	if ((cntfile->filename = malloc (strlen (filename) + 1)) == NULL)
		perr (ERR_MEMALLOC) ;
	strcpy (cntfile->filename, filename) ;

	if ((cntfile->headers = malloc (sizeof *cntfile->headers)) == NULL)
		perr (ERR_MEMALLOC) ;

	readheaders (cntfile) ;

	/* read the file (this fills the rest of cntfile) */
	cntfile->cntbuflhead = initbuflist () ;
	if (noncntbufuse[cntfile->fileformat])
		cntfile->noncntbuflhead = initbuflist () ;
	else
		cntfile->noncntbuflhead = NULL ;

	switch (cntfile->fileformat) {
		case FF_BMP:
		bmp_readfile (cntfile) ;
		break ;

		case FF_WAV:
		wav_readfile (cntfile) ;
		break ;

		case FF_AU:
		au_readfile (cntfile) ;
		break ;
	}

	bufseek (cntfile->cntbuflhead, 0, BUFSEEK_SET) ;
	if (noncntbufuse[cntfile->fileformat])
		bufseek (cntfile->noncntbuflhead, 0, BUFSEEK_SET) ;

	return cntfile ;
}

/* writes the file described in the cntfile structure to disk */
void writecntfile (CNTFILE *cntfile)
{
	switch (cntfile->fileformat) {
		case FF_BMP:
		bmp_writefile (cntfile) ;
		break ;

		case FF_WAV:
		wav_writefile (cntfile) ;
		break ;

		case FF_AU:
		au_writefile (cntfile) ;
		break ;
	}

	bufseek (cntfile->cntbuflhead, 0, BUFSEEK_SET) ;
	if (noncntbufuse[cntfile->fileformat])
		bufseek (cntfile->noncntbuflhead, 0, BUFSEEK_SET) ;

	return ;
}
	
/* reads a plain data file into a PLNFILE structure */
PLNFILE *readplnfile (const char *filename)
{
	PLNFILE *plnfile ;
	int c ;

	if ((plnfile = malloc (sizeof *plnfile)) == NULL)
		perr (ERR_MEMALLOC) ;

	if (strcmp (filename, "-") == 0)
		plnfile->stream = stdin ;
	else {
		if ((plnfile->stream = fopen (filename, "rb")) == NULL)
			perr (ERR_PLNOPEN) ;
	}

	if ((plnfile->filename = malloc (strlen (filename) + 1)) == NULL)
		perr (ERR_MEMALLOC) ;
	strcpy (plnfile->filename, filename) ;

	plnfile->plnbuflhead = initbuflist () ;

	while ((c = getc (plnfile->stream)) != EOF)
		bufsetbyte (c, plnfile->plnbuflhead) ;

	if (ferror (plnfile->stream))
		perr (ERR_PLNREAD) ;

	bufseek (plnfile->plnbuflhead, 0, BUFSEEK_SET) ;

	return plnfile ;
}

/* writes the file described in the plnfile structure to disk */
void writeplnfile (PLNFILE *plnfile)
{
	int c ;
	long i ;

	for (i = 0 ; i < sthdr.nbytesplain ; i++) {
		c = bufgetbyte (plnfile->plnbuflhead) ;
		putc (c, plnfile->stream) ;
	}

	if (ferror (plnfile->stream))
		perr (ERR_PLNWRITE) ;

	bufseek (plnfile->plnbuflhead, 0, BUFSEEK_SET) ;

	return ;
}

int read16_le (FILE *file)
{
	int bytes[2] ;

	bytes[0] = getc (file) ;
	bytes[1] = getc (file) ;

	return ((bytes[1] << 8) | bytes[0]) ;
}

int read16_be (FILE *file)
{
	int bytes[2] ;

	bytes[0] = getc (file) ;
	bytes[1] = getc (file) ;

	return ((bytes[0] << 8) | bytes[1]) ;
}

unsigned long read32_le (FILE *file)
{
	int bytes[4] ;

	bytes[0] = getc (file) ;
	bytes[1] = getc (file) ;
	bytes[2] = getc (file) ;
	bytes[3] = getc (file) ;

	return ((bytes[3] << 24) | (bytes[2] << 16) | (bytes[1] << 8) | bytes[0]) ;
}

unsigned long read32_be (FILE *file)
{
	int bytes[4] ;

	bytes[0] = getc (file) ;
	bytes[1] = getc (file) ;
	bytes[2] = getc (file) ;
	bytes[3] = getc (file) ;

	return ((bytes[0] << 24) | (bytes[1] << 16) | (bytes[2] << 8) | bytes[3]) ;
}

void write16_le (FILE *file, int val)
{
	putc (val & 0xFF, file) ;
	putc ((val >> 8) & 0xFF, file) ;
}

void write16_be (FILE *file, int val)
{
	putc ((val >> 8) & 0xFF, file) ;
	putc (val & 0xFF, file) ;
}

void write32_le (FILE *file, unsigned long val)
{
	putc (val & 0xFF, file) ;
	putc ((val >> 8) & 0xFF, file) ;
	putc ((val >> 16) & 0xFF, file) ;
	putc ((val >> 24) & 0xFF, file) ;
}

void write32_be (FILE *file, unsigned long val)
{
	putc ((val >> 24) & 0xFF, file) ;
	putc ((val >> 16) & 0xFF, file) ;
	putc ((val >> 8) & 0xFF, file) ;
	putc (val & 0xFF, file) ;
}

void cp32ul2uc_be (unsigned char *dest, unsigned long src)
{
	dest[0] = (src >> 24) & 0xFF ;
	dest[1] = (src >> 16) & 0xFF ;
	dest[2] = (src >> 8) & 0xFF ;
	dest[3] = src & 0xFF ;
}

void cp32uc2ul_be (unsigned long *dest, unsigned char *src)
{
	*dest = src[0] << 24 ;
	*dest += src[1] << 16 ;
	*dest += src[2] << 8 ;
	*dest += src[3] ;
}

void cp32ul2uc_le (unsigned char *dest, unsigned long src)
{
	dest[0] = src & 0xFF ;
	dest[1] = (src >> 8) & 0xFF ;
	dest[2] = (src >> 16) & 0xFF ;
	dest[3] = (src >> 24) & 0xFF ;
}

void cp32uc2ul_le (unsigned long *dest, unsigned char *src)
{
	*dest = src[0] ;
	*dest += src[1] << 8 ;
	*dest += src[2] << 16 ;
	*dest += src[3] << 24 ;
}
