/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "steghide.h"
#include "md5.h"

int blocks = 0 ;

static unsigned long md5_F (unsigned long x, unsigned long y, unsigned long z) ;
static unsigned long md5_G (unsigned long x, unsigned long y, unsigned long z) ;
static unsigned long md5_H (unsigned long x, unsigned long y, unsigned long z) ;
static unsigned long md5_I (unsigned long x, unsigned long y, unsigned long z) ;
static unsigned long rotleft (unsigned long x, int n) ;
static unsigned long md5_FF (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s) ;
static unsigned long md5_GG (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s) ;
static unsigned long md5_HH (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s) ;
static unsigned long md5_II (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s) ;
static unsigned long *md5_append (char *passphrase) ;
static unsigned char *md5_calculate (unsigned long *passphrase) ;

static unsigned long md5_F (unsigned long x, unsigned long y, unsigned long z)
{
	return ((x & y) | (~x & z)) ;
}

static unsigned long md5_G (unsigned long x, unsigned long y, unsigned long z)
{
	return ((x & z) | (y & ~z)) ;
}

static unsigned long md5_H (unsigned long x, unsigned long y, unsigned long z)
{
	return (x ^ y ^ z) ;
}

static unsigned long md5_I (unsigned long x, unsigned long y, unsigned long z)
{
	return (y ^ (x | ~z)) ;
}

static unsigned long rotleft (unsigned long x, int n)
{
	return ((x << n) | (x >> (32 - n))) ;
}

static unsigned long md5_FF (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s)
{
	return (b + rotleft ((a + md5_F (b, c, d) + x + t), s)) ;
}

static unsigned long md5_GG (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s)
{
	return (b + rotleft ((a + md5_G (b, c, d) + x + t), s)) ;
}

static unsigned long md5_HH (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s)
{
	return (b + rotleft ((a + md5_H (b, c, d) + x + t), s)) ;
}

static unsigned long md5_II (unsigned long a, unsigned long b, unsigned long c, unsigned long d, unsigned long x, unsigned long t, int s)
{
	return (b + rotleft ((a + md5_I (b, c, d) + x + t), s)) ;
}

/* prepares passphrase for hashing and returns result */
static unsigned long *md5_append (char *passphrase)
{
	int i, j ;
	unsigned long reslen, len, *retval ;
	char *tmp ;

	len = strlen (passphrase) ;

	blocks = ((len + 8) / 64) + 1 ;
	if ((tmp = malloc (blocks * 64)) == NULL)
		perr (ERR_MEMALLOC) ;

	strcpy (tmp, passphrase) ;

	i = len ;

	tmp[i++] = '\200' ;
	while (((i + 8) % 64) != 0)
		tmp[i++] = '\0' ;

	if ((retval = malloc (i + 8)) == NULL)
		perr (ERR_MEMALLOC) ;

	reslen = i ;
	for (j = 0 ; j < (reslen / 4) ; j++)
		cp32uc2ul_le (retval, (unsigned char *) tmp) ;

	retval[j++] = len * 8 ;
	retval[j] = 0 ;

	free (tmp) ;

	return retval ;
}

/* calculates MD5 checksum of passphrase */
static unsigned char *md5_calculate (unsigned long *passphrase)
{
	unsigned long A, B, C, D, AA, BB, CC, DD, X[16] ;
	int i, j ;
	unsigned char *retval ;

	A = 0x67452301UL ;
	B = 0xEFCDAB89UL ;
	C = 0x98BADCFEUL ;
	D = 0x10325476UL ;

	for (i = 0 ; i < blocks ; i++) {
		for (j = 0 ; j < 16 ; j++)
			X[j] = passphrase[i * 16 + j] ;

		AA = A ;
		BB = B ;
		CC = C ;
		DD = D ;

		/* 1st round */
		A = md5_FF (A, B, C, D, X[md5_k[0][0]], md5_T[0][0], md5_s[0][0]) ;
		D = md5_FF (D, A, B, C, X[md5_k[0][1]], md5_T[0][1], md5_s[0][1]) ;
		C = md5_FF (C, D, A, B, X[md5_k[0][2]], md5_T[0][2], md5_s[0][2]) ;
		B = md5_FF (B, C, D, A, X[md5_k[0][3]], md5_T[0][3], md5_s[0][3]) ;
		A = md5_FF (A, B, C, D, X[md5_k[0][4]], md5_T[0][4], md5_s[0][4]) ;
		D = md5_FF (D, A, B, C, X[md5_k[0][5]], md5_T[0][5], md5_s[0][5]) ;
		C = md5_FF (C, D, A, B, X[md5_k[0][6]], md5_T[0][6], md5_s[0][6]) ;
		B = md5_FF (B, C, D, A, X[md5_k[0][7]], md5_T[0][7], md5_s[0][7]) ;
		A = md5_FF (A, B, C, D, X[md5_k[0][8]], md5_T[0][8], md5_s[0][8]) ;
		D = md5_FF (D, A, B, C, X[md5_k[0][9]], md5_T[0][9], md5_s[0][9]) ;
		C = md5_FF (C, D, A, B, X[md5_k[0][10]], md5_T[0][10], md5_s[0][10]) ;
		B = md5_FF (B, C, D, A, X[md5_k[0][11]], md5_T[0][11], md5_s[0][11]) ;
		A = md5_FF (A, B, C, D, X[md5_k[0][12]], md5_T[0][12], md5_s[0][12]) ;
		D = md5_FF (D, A, B, C, X[md5_k[0][13]], md5_T[0][13], md5_s[0][13]) ;
		C = md5_FF (C, D, A, B, X[md5_k[0][14]], md5_T[0][14], md5_s[0][14]) ;
		B = md5_FF (B, C, D, A, X[md5_k[0][15]], md5_T[0][15], md5_s[0][15]) ;

		/* 2nd round */
		A = md5_GG (A, B, C, D, X[md5_k[1][0]], md5_T[1][0], md5_s[1][0]) ;
		D = md5_GG (D, A, B, C, X[md5_k[1][1]], md5_T[1][1], md5_s[1][1]) ;
		C = md5_GG (C, D, A, B, X[md5_k[1][2]], md5_T[1][2], md5_s[1][2]) ;
		B = md5_GG (B, C, D, A, X[md5_k[1][3]], md5_T[1][3], md5_s[1][3]) ;
		A = md5_GG (A, B, C, D, X[md5_k[1][4]], md5_T[1][4], md5_s[1][4]) ;
		D = md5_GG (D, A, B, C, X[md5_k[1][5]], md5_T[1][5], md5_s[1][5]) ;
		C = md5_GG (C, D, A, B, X[md5_k[1][6]], md5_T[1][6], md5_s[1][6]) ;
		B = md5_GG (B, C, D, A, X[md5_k[1][7]], md5_T[1][7], md5_s[1][7]) ;
		A = md5_GG (A, B, C, D, X[md5_k[1][8]], md5_T[1][8], md5_s[1][8]) ;
		D = md5_GG (D, A, B, C, X[md5_k[1][9]], md5_T[1][9], md5_s[1][9]) ;
		C = md5_GG (C, D, A, B, X[md5_k[1][10]], md5_T[1][10], md5_s[1][10]) ;
		B = md5_GG (B, C, D, A, X[md5_k[1][11]], md5_T[1][11], md5_s[1][11]) ;
		A = md5_GG (A, B, C, D, X[md5_k[1][12]], md5_T[1][12], md5_s[1][12]) ;
		D = md5_GG (D, A, B, C, X[md5_k[1][13]], md5_T[1][13], md5_s[1][13]) ;
		C = md5_GG (C, D, A, B, X[md5_k[1][14]], md5_T[1][14], md5_s[1][14]) ;
		B = md5_GG (B, C, D, A, X[md5_k[1][15]], md5_T[1][15], md5_s[1][15]) ;

		/* 3rd round */
		A = md5_HH (A, B, C, D, X[md5_k[2][0]], md5_T[2][0], md5_s[2][0]) ;
		D = md5_HH (D, A, B, C, X[md5_k[2][1]], md5_T[2][1], md5_s[2][1]) ;
		C = md5_HH (C, D, A, B, X[md5_k[2][2]], md5_T[2][2], md5_s[2][2]) ;
		B = md5_HH (B, C, D, A, X[md5_k[2][3]], md5_T[2][3], md5_s[2][3]) ;
		A = md5_HH (A, B, C, D, X[md5_k[2][4]], md5_T[2][4], md5_s[2][4]) ;
		D = md5_HH (D, A, B, C, X[md5_k[2][5]], md5_T[2][5], md5_s[2][5]) ;
		C = md5_HH (C, D, A, B, X[md5_k[2][6]], md5_T[2][6], md5_s[2][6]) ;
		B = md5_HH (B, C, D, A, X[md5_k[2][7]], md5_T[2][7], md5_s[2][7]) ;
		A = md5_HH (A, B, C, D, X[md5_k[2][8]], md5_T[2][8], md5_s[2][8]) ;
		D = md5_HH (D, A, B, C, X[md5_k[2][9]], md5_T[2][9], md5_s[2][9]) ;
		C = md5_HH (C, D, A, B, X[md5_k[2][10]], md5_T[2][10], md5_s[2][10]) ;
		B = md5_HH (B, C, D, A, X[md5_k[2][11]], md5_T[2][11], md5_s[2][11]) ;
		A = md5_HH (A, B, C, D, X[md5_k[2][12]], md5_T[2][12], md5_s[2][12]) ;
		D = md5_HH (D, A, B, C, X[md5_k[2][13]], md5_T[2][13], md5_s[2][13]) ;
		C = md5_HH (C, D, A, B, X[md5_k[2][14]], md5_T[2][14], md5_s[2][14]) ;
		B = md5_HH (B, C, D, A, X[md5_k[2][15]], md5_T[2][15], md5_s[2][15]) ;

		/* 4th round */
		A = md5_II (A, B, C, D, X[md5_k[3][0]], md5_T[3][0], md5_s[3][0]) ;
		D = md5_II (D, A, B, C, X[md5_k[3][1]], md5_T[3][1], md5_s[3][1]) ;
		C = md5_II (C, D, A, B, X[md5_k[3][2]], md5_T[3][2], md5_s[3][2]) ;
		B = md5_II (B, C, D, A, X[md5_k[3][3]], md5_T[3][3], md5_s[3][3]) ;
		A = md5_II (A, B, C, D, X[md5_k[3][4]], md5_T[3][4], md5_s[3][4]) ;
		D = md5_II (D, A, B, C, X[md5_k[3][5]], md5_T[3][5], md5_s[3][5]) ;
		C = md5_II (C, D, A, B, X[md5_k[3][6]], md5_T[3][6], md5_s[3][6]) ;
		B = md5_II (B, C, D, A, X[md5_k[3][7]], md5_T[3][7], md5_s[3][7]) ;
		A = md5_II (A, B, C, D, X[md5_k[3][8]], md5_T[3][8], md5_s[3][8]) ;
		D = md5_II (D, A, B, C, X[md5_k[3][9]], md5_T[3][9], md5_s[3][9]) ;
		C = md5_II (C, D, A, B, X[md5_k[3][10]], md5_T[3][10], md5_s[3][10]) ;
		B = md5_II (B, C, D, A, X[md5_k[3][11]], md5_T[3][11], md5_s[3][11]) ;
		A = md5_II (A, B, C, D, X[md5_k[3][12]], md5_T[3][12], md5_s[3][12]) ;
		D = md5_II (D, A, B, C, X[md5_k[3][13]], md5_T[3][13], md5_s[3][13]) ;
		C = md5_II (C, D, A, B, X[md5_k[3][14]], md5_T[3][14], md5_s[3][14]) ;
		B = md5_II (B, C, D, A, X[md5_k[3][15]], md5_T[3][15], md5_s[3][15]) ;

		A += AA ;
		B += BB ;
		C += CC ;
		D += DD ;
	}

	if ((retval = malloc (16)) == NULL)
		perr (ERR_MEMALLOC) ;

	cp32ul2uc_le (&retval[0], A) ;
	cp32ul2uc_le (&retval[4], B) ;
	cp32ul2uc_le (&retval[8], C) ;
	cp32ul2uc_le (&retval[12], D) ;

	return retval ;
}

unsigned char *md5_generatekey (char *passphrase)
{
	unsigned long *passprep ;
	unsigned char *retval ;

	passprep = md5_append (passphrase) ;
	retval = md5_calculate (passprep) ;
	free (passprep) ;

	return retval ;
}
