/*
 * steghide 0.3-1 - a steganography tool
 * Copyright (C) 1999 Stefan Hetzl <shetzl@teleweb.at>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#ifndef SH_STEGHIDE_H
#define SH_STEGHIDE_H

/* CNTFILE/PLNFILE structures and substructures */
typedef struct struct_BUFFER {
	/* pointer to the data of this BUFFER structure */
	void	*ptr ;
	/* position of the current byte (0-[n-1]) */
	long	bytepos ;
	/* position of the current bit (0-7) */
	int		bitpos ;
	/* number of allocated bytes that are filled with data */
	long	nbytes ;
	/* number of bytes that have been allocated */
	long	nbytesalloc ;
	/* number of the current buffer in linked list (is only set in 1st element) */
	int		curbufnum ;
	/* total numbers of buffer that point to data in linked list (is only set in 1st element) */
	int		nbufs ;
	/* pointer to previous buffer in linked list ((x->prev == x) indicates first element) */
	struct struct_BUFFER	*prev ;
	/* pointer to next buffer in linked list ((x->next == x) indicates last element) */
	struct struct_BUFFER	*next ;
} BUFFER ;

/* bmp header structures */
typedef struct struct_BITMAPFILEHEADER {
	unsigned short	bfType ;
	unsigned long	bfSize ;
	unsigned short	bfReserved1 ;
	unsigned short	bfReserved2 ;
	unsigned long	bfOffBits ;
} BITMAPFILEHEADER ;

typedef struct struct_BITMAPXHEADER { /* stores data from BITMAPINFO- and BITMAPCORE- HEADER */
	unsigned long	Size ;
	signed long		Width ;
	signed long		Height ;
	unsigned short	Planes ;
	unsigned short	BitCount ;
	unsigned long	Compression ;
	unsigned long	SizeImage ;
	signed long		XPelsPerMeter ;
	signed long		YPelsPerMeter ;
	unsigned long	ClrUsed ;
	unsigned long	ClrImportant ;
} BITMAPXHEADER ;

typedef struct struct_BMPHEADERS {
	BITMAPFILEHEADER	bmfh ;
	BITMAPXHEADER		bmxh ;
} BMPHEADERS ;

/* wav header structures */
typedef struct struct_CHUNKHEADER {
	char			id[4] ;
	unsigned long	len ;
} CHUNKHEADER ;

typedef struct struct_FMTCHUNK_PCM {
	signed short	FormatTag ;
	unsigned short	Channels ;
	unsigned long	SamplesPerSec ;
	unsigned long	AvgBytesPerSec ;
	unsigned short	BlockAlign ;
	unsigned short	BitsPerSample ;
} FMTCHUNK_PCM ;

typedef struct struct_WAVHEADERS {
	CHUNKHEADER		riffchhdr ;
	char			id_wave[4] ;
	CHUNKHEADER		fmtchhdr ;
	FMTCHUNK_PCM	fmtch ;
	CHUNKHEADER		datachhdr ;
} WAVHEADERS ;
	
/* au header structures */
typedef struct struct_AUHEADERS {
	char			id[4] ;
	unsigned long	offset ;
	unsigned long	size ;
	unsigned long	encoding ;
	unsigned long	samplerate ;
	unsigned long	channels ;
} AUHEADERS ;

/* CNTHEADERS contains the headers of all supported file formats */
typedef union union_CNTHEADERS {
	BMPHEADERS		bmp ;
	WAVHEADERS		wav ;
	AUHEADERS		au ;
} CNTHEADERS ;

/* CNTFILE describes a containerdata file */
typedef struct struct_CNTFILE {
	/* the file's data stream */
	FILE		*stream ;
	/* name of the file */
	char		*filename ;
	/* fileformat of the file (one of the FF_* constants) */
	int			fileformat ;
	/* headers of the file */
	CNTHEADERS	*headers ;
	/* pointer to data of unsupported header structures (legal in fileformat, but not needed by steghide)
	   those are saved here to preserve all data in headers and the original filesize when embedding data */
	void		*unsuphdr ;
	/* length of data of unsupported header structures (in bytes) */
	int			unsuphdrlen ;
	/* first element of linked list of buffers that contain the containerbytes of the file */
	BUFFER		*cntbuflhead ;
	/* first element of linked list of buffers that contain all other bytes of the file
	   these buffers contain all bytes in which data must not be hidden */
	BUFFER		*noncntbuflhead ;
} CNTFILE ;

/* PLNFILE describes a plaindata file */
typedef struct struct_PLNFILE {	
	/* file as data stream */
	FILE	*stream ;
	/* name of file */
	char	*filename ;
	/* first element of linked list of buffers that contain the data of the file */
	BUFFER	*plnbuflhead ;
} PLNFILE ;

/* STEGOHEADER structure and substructures */
typedef struct struct_SPCMTDINFO_CNST {
	/* the constant space between two stegobytes */
	unsigned char	space ;
} SPCMTDINFO_CNST ;

typedef struct struct_SPCMTDINFO_LINC {
	/* seed for random number generator */
	unsigned long	seed ;
	/* largest value the random number generator will create */
	unsigned char	maxspace ;
} SPCMTDINFO_LINC ;

typedef union union_SPCMTDINFO {
	SPCMTDINFO_CNST	cnst ;
	SPCMTDINFO_LINC	linc ;
} SPCMTDINFO ;

/* this structure describes the plaindata and how it is embedded */
typedef struct struct_STEGOHEADER {
	/* spacemtd to calculate space between stegobytes (one of the MTD_* constants) */
	unsigned char	spacemtd ;
	/* spacemtd-specific parameters */
	SPCMTDINFO		spcmtdinfo ;
	/* mask that is used to determine bits that contain data
	   if stgmask is for example 1, only the least significant bit will be used */
	unsigned char	stgmask ;
	/* length of the plaindata (in bytes) */
	unsigned long	nbytesplain ;
} STEGOHEADER ;

/* size of stego header components when embedded in stego file */
#define SIZE_SPACEMTD			1
#define SIZE_SPCMTDINFO_CNST	1
#define SIZE_SPCMTDINFO_LINC	5
#define SIZE_ENCRYPTION			1
#define SIZE_STGMASK			1
#define SIZE_NBYTESPLAIN		4
#define MAXSIZE_STHDR			SIZE_SPACEMTD + SIZE_SPCMTDINFO_LINC + SIZE_ENCRYPTION + SIZE_STGMASK + SIZE_NBYTESPLAIN

/* misc. constants used in/for the stegoheader */
#define SPCMTD_CNST	1
#define SPCMTD_LINC	2
#define STHDR_BLOCKS_BLOWFISH	2

/* constants that indicate the container file format */
#define FF_UNKNOWN	0
#define FF_BMP		1
#define FF_WAV		2
#define FF_AU		3

/* encryption */
#define ENC_NONE		0
#define ENC_BLOWFISH	1

/* size of one buffer in linked list */
#define BUFSIZE		10240

/* values for whence-parameter in buflseek */
#define BUFSEEK_SET		0
#define BUFSEEK_CUR		1
#define BUFSEEK_END		2
#define ENDOFBUF		-1

/* is used for parameter freesubstructs of cleanupcntfile */
#define FSS_NO	0
#define FSS_YES	1

/* values for action-parameter in spacetonextbyte() etc. */
#define SP_RESET		1
#define SP_GETVALUE		2

/* used by the LINC-spacemtd */
#define LINC_b		31415821UL
#define LINC_m		100000000UL
#define LINC_m1		10000

/* bmp file format */
#define BMP_SIZE_BMFILEHEADER	14
#define BMP_SIZE_BMINFOHEADER	40
#define BMP_SIZE_BMCOREHEADER	12
#define BMP_ID_BM				19778
#define BMP_BI_RGB				0

/* wav file format */
#define WAV_FORMAT_PCM		1
#define WAV_SIZE_FMTCHUNK	16

/* au file format */
#define AU_SIZE_HEADER		24

/* error constants */
#define ERR_OK					0	/* no error */

#define ERR_WRONGARG			1	/* wrong argument syntax */

#define ERR_CNTOPEN				2	/* can not open container file */
#define ERR_STGOPEN				3	/* can not open stego file */
#define ERR_PLNOPEN				4	/* can not open plain file */
#define ERR_CNTREAD				5	/* can not read from container file */
#define ERR_STGREAD				6	/* can not read from stego file */
#define ERR_PLNREAD				7	/* can not read from plain file */
#define ERR_STGWRITE			8	/* can not write to stego file */
#define ERR_PLNWRITE			9	/* can not write to plain file */
#define ERR_CNTCLOSE			10	/* can not close container file */
#define ERR_STGCLOSE			11	/* can not close stego file */
#define ERR_PLNCLOSE			12	/* can not close plain file */

#define ERR_CNTUNKNOWNFFORMAT	13	/* container file has unknown file format */
#define ERR_STGUNKNOWNFFORMAT	14	/* stego file has unknown file format */
#define ERR_CNTUNSUPBMPFILE		15	/* container file is unsupported bmp file (not win3.x nor OS/2 1.x) */
#define ERR_STGUNSUPBMPFILE		16	/* stego file is unsupported bmp file */
#define ERR_CNTBMPWITHCOMPR		17	/* container file is bmp file that uses compression */
#define ERR_STGBMPWITHCOMPR		18	/* stego file is bmp file that uses compression */
#define ERR_CNTUNSUPWAVFILE		19	/* container file is unsupported wav file (not PCM) */
#define ERR_STGUNSUPWAVFILE		20	/* stego fileis unsupported wav file */
#define ERR_CNTBPSNOTMULTOF8	21	/* container file is wav file with bits/sample rate not a multiple of eight */
#define ERR_STGBPSNOTMULTOF8	22	/* stego file is wav file with bits/sample rate not a multiple of eight */

#define ERR_MEMALLOC			23	/* can not allocate memory */
#define ERR_CNTTOOSHORT			24	/* container file is too short to embed plain data there */
#define ERR_STGTOOSHORT			25	/* stego file is too short to extract plain data from there */

/* program name */
#define PROGNAME	"steghide"

/* bufmanag.c */
extern int noncntbufuse[] ;

CNTFILE *createstgfile (CNTFILE *cntfile, const char *stgfilename) ;
PLNFILE *createplnfile (const char *plnfilename) ;
void cleanupcntfile (CNTFILE *cntfile, int freesubstructs) ;
void cleanupplnfile (PLNFILE *plnfile) ;
BUFFER *initbuflist (void) ;
BUFFER *getcurbuf (BUFFER *buflhead) ;
int bufendofbuf (BUFFER *buflhead) ;
unsigned long bufget32_be (BUFFER *buflhead) ;
void bufset32_be (unsigned long val, BUFFER *buflhead) ;
int bufgetbyte (BUFFER *buflhead) ;
void bufsetbyte (int byteval, BUFFER *buflhead) ;
int bufgetbit (BUFFER *buflhead) ;
void bufsetbit (int bitval, BUFFER *buflhead) ;
void bufsetbitpos (int bitpos, BUFFER *buflhead) ;
int bufgetbitpos (BUFFER *buflhead) ;
int bufseek (BUFFER *buflhead, long offset, int whence) ;
long buftell (BUFFER *buflhead) ;
long bufgetdatalen (BUFFER *buflhead) ;
void buffree (BUFFER *buflhead) ;

/* crypto.c */
void encrypt_data (PLNFILE *plnfile, int encryption, char *passphrase) ;
void decrypt_data (PLNFILE *plnfile, int encryption, char *passphrase) ;
void encrypt_sthdr (unsigned char *bytes, int len, int encryption, char *passphrase) ;
void decrypt_sthdr (unsigned char *bytes, int encryption, char *passphrase) ;

/* fformats.c */
void readheaders (CNTFILE *file) ;
void bmp_readfile (CNTFILE *file) ;
void bmp_writefile (CNTFILE *file) ;
void wav_readfile (CNTFILE *file) ;
void wav_writefile (CNTFILE *file) ;
void au_readfile (CNTFILE *file) ;
void au_writefile (CNTFILE *file) ;

/* io.c */
CNTFILE *readcntfile (const char *filename) ;
void writecntfile (CNTFILE *cntfile) ;
PLNFILE *readplnfile (const char *filename) ;
void writeplnfile (PLNFILE *plnfile) ;
int read16_le (FILE *file) ;
int read16_be (FILE *file) ;
unsigned long read32_le (FILE *file) ;
unsigned long read32_be (FILE *file) ;
void write16_le (FILE *file, int val) ;
void write16_be (FILE *file, int val) ;
void write32_le (FILE *file, unsigned long val) ;
void write32_be (FILE *file, unsigned long val) ;
void cp32ul2uc_be (unsigned char *dest, unsigned long src) ;
void cp32uc2ul_be (unsigned long *dest, unsigned char *src) ;
void cp32ul2uc_le (unsigned char *dest, unsigned long src) ;
void cp32uc2ul_le (unsigned long *dest, unsigned char *src) ;

/* main.c */
#define ACTN_EMBED		1
#define ACTN_EXTRACT	2

extern int args_action ;
extern int args_space_sthdr ;
extern int args_encryption_sthdr ;
extern char *args_passphrase_sthdr ;
extern int args_encryption_data ;
extern char *args_passphrase_data ;
extern char *args_fn_cnt ;
extern char *args_fn_stg ;
extern char *args_fn_pln ;

void perr (int errn) ;

/* md5.c */
unsigned char *md5_generatekey (char *passphrase) ;

/* stegano.c */
extern STEGOHEADER sthdr ;

void embeddata (BUFFER *cntbuflhead, BUFFER *plnbuflhead) ;
BUFFER *extractdata (BUFFER *stgbuflhead) ;
void writesthdr (BUFFER *cntbuflhead, int encryption, char *passphrase) ;
void readsthdr (BUFFER *stgbuflhead, int encryption, char *passphrase) ;

#endif	/* not defined SH_STEGHIDE_H */
